#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include "RTjpegN.h"
#include "XJ.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include "minilzo.h"
#include "yuv2rgb.h"
#include "nuppelvideo.h"
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include "areaDeinterlace.h"
#include "rtjpeg_plugin.h"

#define MAXPREBUFFERS 60

// -- lzo static vars ------------------------------------
#define IN_LEN          (128*1024L)
#define OUT_LEN         (IN_LEN + IN_LEN / 64 + 16 + 3)
// -- lzo static vars end --------------------------------

int compoff = 0;
extern int editmode;


static void hup(int foo)
{
 XJ_exit();
 exit(0);
}

void (*vidconv)(__u8 *buf, __u8 *rgb, int stride);

int initsound()
{
  int audio;
// int fmt;
  int bits=16, stereo=1, speed=44100;

  audio = open("/dev/audio", O_WRONLY);
  if (audio == -1) {
    fprintf(stderr,"%s\n", "can't open audio device!");
    return(0);
  }

  // linux audio seems to have allways 64k buffer
  
  // if (ioctl(audio, SNDCTL_DSP_GETFMTS, &fmt) < 0) {
  //   fprintf(stderr,"%s\n", "cant get audio fmt's!");
  //   exit(3);
  // }

  if (ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &bits) < 0) {
    fprintf(stderr,"%s\n", "can't open audio device(SNDCTL_DSP_SAMPLESIZE)!");
    return(0);
  }

  if (ioctl(audio, SNDCTL_DSP_STEREO, &stereo) < 0) {
    fprintf(stderr,"%s\n", "can't open audio device(SNDCTL_DSP_STEREO)!");
    return(0);
  }

  if (ioctl(audio, SNDCTL_DSP_SPEED, &speed) < 0) {
    fprintf(stderr,"%s\n", "can't open audio device(SNDCTL_DSP_SPEED)!");
    return(0);
  }

  return(audio);
}

void write_audio(int audio, unsigned char *aubuf, int size)
{
  unsigned char *tmpbuf;
  int written=0, lw=0;

  tmpbuf = aubuf;

  while ( (written < size) && ((lw = write(audio, tmpbuf, size - written)) > 0) )
  {
    if (lw == -1) {
      fprintf(stderr,"%s\n", "can't write to audio device!");
      exit(1);
    }
    written += lw;
    tmpbuf += lw;
  }

  // now sync audio
  // if (ioctl (audio, SNDCTL_DSP_SYNC, NULL) < 0) {
  //   fprintf(stderr,"%s\n", "can't sync audio device!");
  //   return;
  // }
}

// ----------------------------------------------------------
// -- USAGE -------------------------------------------------

void usage()
{
   fprintf(stderr, "\nNuppelVideo Player v0.52         (c)Roman HOCHLEITNER\n");
   fprintf(stderr, "Usage: nuvplay [options] filename \n\n");
   fprintf(stderr, "options: -w n .......... wait n miliseconds (per frame)\n");
   fprintf(stderr, "         -e ............ export audio to stdout\n");
   fprintf(stderr, "         -E filename ... export audio to filename\n");
   fprintf(stderr, "         -a n [msec].... audio-video delay (experimental) [0]\n");
   fprintf(stderr, "         -F n [Hz]...... force eff-DSP frequency [calc.]\n");
   fprintf(stderr, "         -b n .......... videoframes delayed (sync) 0..60 [8]\n");
   fprintf(stderr, "         -f ............ play fast (every 7th fr. no audio)\n");
//   fprintf(stderr, "         -C ............ display changes\n");
   fprintf(stderr, "         -d ............ deinterlace (very slow)\n");
   fprintf(stderr, "         -s ............ do not skip frames\n");
   fprintf(stderr, "         -c ............ only count video frames\n");
   fprintf(stderr, "         -V ............ video only (i.e. no audio)\n");
   fprintf(stderr, "         -Z ............ edit mode (for nuvedit)\n");
   fprintf(stderr, "         -h ............ this help\n\n");
   fprintf(stderr, "Note: this 'player' is far from perfect and if video/audio is not\n");
   fprintf(stderr, "      in sync it's rather more a problem of the player and not of the\n"); 
   fprintf(stderr, "      recordtool, in this case try -b 1 to 30 or -a -200 .. +200\n");
   fprintf(stderr, "      to compensate the delay\n");
   fprintf(stderr, "\n");
   exit(-1);
}


int main(int argc, char *argv[])
{
  int w=384, h=288;
  struct timeval now, last;
  struct timezone tzone;
  //struct rtframeheader *frameheader;
  // int r, out_len;
  int audf=0;
  char c;
  unsigned char *buf, *buf3;
  // unsigned char *buf2, *orig;
  char *sbuf=NULL;
  // unsigned long int tbls[128];
  // __s8 *strm;
  long tf=0;
  yuv2rgb_fun yuvconvert;
  int usecs=0;
  // use count prebuffers to adjust audio/video sync
  int usepre=8;
  int playaudio=1;
   int displaychanges=0; // is currently unsupported, will prob. work again
  int playfast=0;
  int countonly=0;
  int writeaudiotofile=0;
  int deinterlace=0;
  char *audiofilename = "audio";
  char *filename = "";
  unsigned char *audiodata;
  int timecode; 
 
  int prebuffered=0;
  int actpre=0;
  int fnum=0;
  int audiolen;
  char command[1024];
  int playmode=0;
  int skipframes=1;
  int effdsp=-1;
  //struct termios termios;

  unsigned char *prebuf[MAXPREBUFFERS];

  rtjpeg_resample=0; 
  editmode = 0;

  //while ((c=getopt(argc,argv,"w:E:b:dfeCcVh")) != -1) 

  while ((c=getopt(argc,argv,"w:E:b:dF:a:fsecVZh")) != -1) {
    switch(c) {
      case 'w': playfast=atoi(optarg); if(playfast<=1) playfast=2;  playaudio=usepre=0; break;
      case 'E': writeaudiotofile=1; rtjpeg_resample=1;  audiofilename=optarg; break;
      case 'b': usepre=atoi(optarg); playaudio=1; break;
      case 'a': rtjpeg_audiodelay=atoi(optarg); break;
      case 'F': effdsp=atoi(optarg); break;
      case 'f': playfast=1; playaudio=usepre=0; break;
      case 'e': writeaudiotofile=2; rtjpeg_resample=1; playaudio=1; break;
      case 'C': displaychanges=1; break;
      case 'd': deinterlace=1; break;
      case 's': skipframes=0; break;
      case 'c': countonly=1; displaychanges=playaudio=usepre=0; ; break;
      case 'V': playaudio=0; break;
      case 'Z': editmode=1; rtjpeg_audiodelay=-0.20 ; skipframes=0; usepre=0; break;
      case 'h': usage();         break;
 
      default: usage();
    }
  }

  if (optind==argc) usage();
  else filename=argv[optind];


 tzone.tz_minuteswest=-60;
 tzone.tz_dsttime=0;
 now.tv_sec=0; // reset

 // init compression lzo ------------------------------
 if (lzo_init() != LZO_E_OK)
 {
   fprintf(stderr,"%s\n", "lzo_init() failed !!!");
   exit(3);
 }

 signal(SIGINT, &hup);
 
 // now we open the file

 rtjpeg_open(filename);

 w = rtjpeg_video_width;
 h = rtjpeg_video_height;

 if (effdsp != -1) {
   rtjpeg_effdsp = effdsp*100;
 }

 fprintf(stderr, "nuppelvideo file with w=%d and h=%d and fps=%f"
                 "\nframes=%d realframes=%d effdsp=%5.2f\n",
                  rtjpeg_video_width,
                  rtjpeg_video_height,
                  rtjpeg_video_frame_rate,
                  rtjpeg_fakeframescount,
                  rtjpeg_framescount,
                  (double)rtjpeg_effdsp/100.0);


 if (editmode) {
   int flgs;
   //if (-1==fcntl(0, O_NONBLOCK)) {
   //  fprintf(stdout, "cannot set O_NONBLOCK on stdin\n");
   //  exit(1);
   // }
   // if (tcgetattr (0, &termios ) == 0) {
   //    // termios.c_lflag &= ~(ISIG | ICANON);
   //    /* get input on each char, timeout = 0 */
   //    termios.c_cc[VMIN] = 1;
   //    termios.c_cc[VTIME] = 0;
   // 
   //       if(tcsetattr (0, TCSANOW, &termios ) != 0) {
   //         perror("tcsetattr");
   //       }
   //    } else {
   //      perror("tcgetattr");
   //    }
   
   if (-1==(flgs=fcntl(0, F_GETFL))) {
     fprintf(stdout, "cannot F_GETFL on stdin\n");
     exit(1);
   }

   if (-1==fcntl(0, F_SETFL, flgs | O_NONBLOCK)) {
     fprintf(stdout, "cannot set O_NONBLOCK on stdin\n");
     exit(1);
   }

   fprintf(stdout, "%d %d %3.2f %d %5.2f\n", w, h, rtjpeg_video_frame_rate, 
                   rtjpeg_framescount, (double)rtjpeg_effdsp/100.0); 
   fflush(stdout);
 }
 
 if (rtjpeg_fileheader.audioblocks!=0 && playaudio!=0) {
   if (writeaudiotofile) {
     // write to file, file MUST NOT exist
     audf = (writeaudiotofile==1) ? open(audiofilename, O_EXCL|O_CREAT|O_WRONLY|O_SYNC, 0600) : 1;
     if (audf==-1) {
       if (errno!=EEXIST)
         fprintf(stderr, "cannot write to file '%s', no audio is played\n", audiofilename);
       else 
         fprintf(stderr, "file '%s' exists and will not be overwritten\n", audiofilename);
         
       playaudio=0; // reset flag
     }
   } else {
     // write to sound card
     audf=initsound();
     if (audf==0) {
       fprintf(stderr, "Cannot write to audio device, no audio is played\n" );
       playaudio=0; // reset flag
     }
   }
 } else {
   playaudio=0;
 }

// // first frame has to be the Compression "D"ata frame
// read(f, &frameheader, sizeof(frameheader));
// if (frameheader.frametype != 'D') {
//   fprintf(stderr, "nuvplay: Error: Illegal File Format, no Compression(D)ata frame\n" );
//   exit(1);
// }
// read(f, tbls, frameheader.packetlength);
// RTjpeg_init_decompress(tbls, w, h);
 
// strm=malloc(w*h+(w*h)/2);

 if (!countonly && writeaudiotofile==0) {
   sbuf=XJ_init(w, h, "NuppelVideo Player (experimental)", "NuppelVideo");
   switch (XJ_depth)
   {
    case 8: vidconv=RTjpeg_yuvrgb8;
            break;
    case 16: vidconv=RTjpeg_yuvrgb16;
             break;
    case 24: vidconv=RTjpeg_yuvrgb24;
             // break;
    case 32: vidconv=RTjpeg_yuvrgb32; break;
             break;
    default: fprintf(stderr, "Unsupported display depth\n");
             exit(-1);
             break;
   }
 }

// buf=malloc(w*h+(w*h)/2);
// buf2=malloc(w*h+(w*h)/2);
// orig=malloc(w*h+(w*h)/2);

 buf3=malloc(w*h+(w*h)/2);

if (0) {
  int uuu;
  // fill pipe/kernel buffer
  memset(buf,0,16384);
  for (uuu=0;uuu<3;uuu++) {
    write(audf,buf,16384);
  }
}
 
// sleep(1);

 if (XJ_depth==24) XJ_depth=32; 
 yuvconvert=yuv2rgb_init_mmx(XJ_depth, MODE_RGB);

 if (countonly) { 
   fprintf(stderr, "total frames %d\n", rtjpeg_fakeframescount);
   exit(0);
 }

 if (playaudio==0) fprintf(stderr, "playing no audio\n");

 while(!rtjpeg_eof)
 {
  if (writeaudiotofile) {
    buf = rtjpeg_get_frame(fnum, &timecode, -1, &audiodata, &audiolen); // no video
  } else {
    buf = rtjpeg_get_frame(fnum, &timecode, playaudio==0, &audiodata, &audiolen);
  }
  if (rtjpeg_eof) continue;
  if (audiolen!=0) {
    if (writeaudiotofile==1) {
      write(audf, audiodata, audiolen);
      fsync(audf); // play the sound immediately
    } else if (writeaudiotofile>=2) {
      fwrite(audiodata, audiolen, 1, stdout);
      fflush(stdout); // play the sound immediately
    } else {
      // write to soundcard
      write_audio(audf, audiodata, audiolen);
    }
  }

  if (writeaudiotofile) {
    fnum++;
    continue; // no video will be displayed
  }

  // we had to remove the "show (C)hanges feature" :-(
  // maybe i'll move the code to rtjpeg_plugin.c FIXLITE

  // now we have to use/fill the pic-cache for syncing
  // because the soundcard needs 64k before playing --> ca 13pics

  if (usepre!=0) {
    if (prebuffered<usepre) {
      prebuf[prebuffered]=malloc(w*h+(w*h)/2);
      memcpy(prebuf[prebuffered], buf, w*h+(w*h)/2);
      prebuffered++;
      fnum++;
      continue;
    } else {
      // place the neweset buffer in the oldest
      memcpy(buf3, prebuf[actpre], w*h+(w*h)/2);
      memcpy(prebuf[actpre], buf, w*h+(w*h)/2);
      actpre++;
      if (actpre>=usepre) actpre=0;
    }
  } else {
    // if we play no audio we don;t need picbuffers for sync
    memcpy(buf3, buf, w*h+(w*h)/2);
  }

  // calaculate the delay

  if (!playfast) {
    if (now.tv_sec==0) {
      gettimeofday(&now, &tzone);
      last=now;
      usecs= (1000000.0/rtjpeg_video_frame_rate)/2;
      if (usecs>0) usleep(usecs);
    } else {
      gettimeofday(&now, &tzone);
      usecs = 1000000.0/rtjpeg_video_frame_rate - 
              ((now.tv_sec-last.tv_sec)*1000000 + now.tv_usec-last.tv_usec);
      last=now;
      //fprintf(stderr, "\nwaiting %d usecs    ", usecs);
      if (usecs > (1000000/rtjpeg_video_frame_rate)/2) usecs= (1000000/rtjpeg_video_frame_rate)/2;
    }
  }

  if (playfast!=1 || tf%7==0) {
    if (!skipframes || usecs > (-1000000/rtjpeg_video_frame_rate)) {
      if (deinterlace) AreaBasedDeinterlacerYUV420(buf3, w, h); 
      if (yuvconvert==NULL) {
        // slow for all cpu's when -DMMX is not defined during compilation
        // non MMX code
        vidconv(buf3, sbuf, 0); 
      } else {
        // slow for mmx cpus (xv or dga [1.0 or 2.0] would be faster ;-)
        // but at least twice as fast as the non-MMX code, try it
        yuvconvert(sbuf, buf3, buf3 + (w*h),  buf3 + (w*h + ((w*h)>>2)), w, h);
      }
      XJ_show(); // display the frame
    } // else skip the displayupdate + yuv->rgb to spare time and maybe sync again
  } else {
    tf++;
    fnum++;
    continue;
  }

  if (editmode) {
    int ret, len, newnum=-1;
   
    if (playmode==1 && fnum==rtjpeg_framescount-1) {
      playmode=0;
    }

    do {
      command[0]='\0';
      if (NULL==fgets(command, 1024, stdin) && feof(stdin)) {
        fprintf(stderr, "got EOF from server, terminating\n");
        exit(1);
      }
      ret = strlen(command);
      if (ret) {
        len = ret;
        while (ret!=0 && command[len-1]!='\n') {
          if (NULL != fgets(command+len, 1024-len, stdin) && feof(stdin)) {
            fprintf(stderr, "got EOF from server, terminating\n");
            exit(1);
          }
          len = strlen(command);
        }
        sscanf(command, "%d %d\n", &playmode, &newnum);
      } else {
        fprintf(stdout, "%d %d\n", fnum, 0); // second would be state ???
        fflush(stdout);
      } 
      if (!playmode) usleep(200000);
    } while (playmode==0 && newnum==-1);

    if (newnum>=0) {
      fnum = newnum;
      continue;
    }
  }

  if (!playfast) {
      if (usecs>0) usleep(usecs);
  }
 
  if (playfast >1) {
    usleep(playfast*1000);
  }  
  tf++;
  fnum++;
 }
 
 fprintf(stderr, "total frames = %d played frames = %ld\n", rtjpeg_fakeframescount, tf+usepre);
 if (!countonly && writeaudiotofile!=2) {
   XJ_exit();
 }

 return 0;
}

